<?php

/*
 * Copyright 2018, 1st Source IT, LLC, EZ Merchant Solutions
 * All rights reserved.
 * Resale prohibited.
 */

namespace Tygh\Api\Entities;

use Tygh\Api\AEntity;
use Tygh\Api\Response;
use Tygh\Registry;

class SquareToken extends AEntity {

	Public function index($id=null, $params=[]) {
		if( empty($id) ) {	// Get all squarepay payment_ids with merchant_id and sq_token
			$data = sq_payment_tokens();
		} else {
			$data = sq_payment_token($id);
		}
		if( empty($data) ) {
			$status = Response::STATUS_NOT_FOUND;
		} else {
			$status = Response::STATUS_OK;
		}
		return array(
			'status' => $status,
			'data' => $data
			);
	}
	
	Public function create($params) {
		return array( 'status' => Response::STATUS_BAD_REQUEST, 'data' => array('error_message'=>'Not supported') );
	}
	
	Public function update($id, $params) {
		$data = sq_payment_token($id);	// Actually processor_params
		
		if( !$data ) 
			return array('status'=>Response::STATUS_NOT_FOUND, 'data' => array('error_message' => "Bad payment_id '$id'") );
		unset($params['payment_id']);
		
		if( empty($params['expires']) ) {
			return array('status'=>Response::STATUS_BAD_REQUEST, array('error_message'=>"No expiration data"));
		}
		if( empty($params['access_token']) ) {
			return array('status'=>Response::STATUS_BAD_REQUEST, array('error_message'=>"No access token"));
		}
		if( empty($params['merchant_id']) ) {
			return array('status'=>Response::STATUS_BAD_REQUEST, array('error_message'=>"No merchant_id"));
		}
		
		$data['token_expires'] = $params['expires'];
		$data['sq_access_token'] = $params['access_token'];
		$data['merchant_id'] = $params['merchant_id'];
		sq_update_payment_token($id, $data);
		return array('status' => Response::STATUS_OK, 'data' => $data);
	}
	
	Public function delete($id) {
		return array( 'status' => Response::STATUS_BAD_REQUEST, 'data' => array() );
	}
	
	Public function privileges() {
		return array(
					 'create' => 'manage_squarepay',
					 'update' => 'manage_squarepay',
					 'delete' => 'manage_squarepay',
					 'index' => 'manage_squarepay'
					 );
	}
}
?>