<?php
/*****************************
* Copyright 2009, 2010, 2011, 2012, 2013 1st Source IT, LLC
* All rights reserved.
* Permission granted for use as
* long as this copyright notice, associated text and
* links remain in tact.
* Licensed for a single domain and a single instance of EZ-cart.
* Additional licenses can be purchased for additonal sites.
*
* http://www.ez-ms.com
* http://www.ez-om.com
* http://www.1sit.com*
*
* End copyright notification
*/
use Tygh\Registry;

if( !class_exists('squarepay_log') ) {
$dir_ar = explode('/', dirname(__FILE__));
$addon_name = array_pop($dir_ar);

define('squarepay_log_logfilename', "/${addon_name}.log");
define('squarepay_log_logdir', dirname(__FILE__));
define('squarepay_log_logfile', squarepay_log_logdir.squarepay_log_logfilename);
define('squarepay_log_end_record', "\r\n");

if( !defined('logAll') ) {
	define('logAll', 0);
	define('logInfo', 1);
	define('logLog', logInfo);
	define('logTrace', 2);
	define('logDetail', 3);
}

class squarepay_log {
	static $buf = array();
	static $output = false;
	static $log_level = 0;
	static $log_area = 'A';
	static $testing = false;
	
	static function init($log_level=0, $output=false, $area = 'A', $display_errors=false) {
		// Too many issues from other addons and core cs-cart to have this be live on other than 
		// test environments....
		if( $display_errors ) {
			error_reporting(E_ALL);
			if( $output )
				ini_set('display_errors', true);
		}

		$err = true;
		if( !is_dir(squarepay_log_logdir) )
			$err = @mkdir(squarepay_log_logdir);
		if( $err )
			register_shutdown_function('squarepay_log_dump_log');
		self::$output = $output;
		self::$log_level = $log_level;
		self::$log_area = $area;
	}
	
	static function add_msg($s, $level=0) {
		if( self::$log_level >= $level )
			self::$buf[] = $s;
	}
	
	static function show_log($entries=10) {
		$raw_log_entries = array_reverse(explode(squarepay_log_end_record, @file_get_contents(squarepay_log_logfile)));
		array_shift($raw_log_entries);		// Last (first) record will always be empty due to entries ending in squarepay_log_end_record
		if( !current($raw_log_entries) )
			$raw_log_entries = array();
		$tot_entries = count($raw_log_entries);
		$entries = ($tot_entries > $entries) ? $entries : $tot_entries;
		echo "<pre>";
//		echo "Logfile: ".squarepay_log_logfile."\n";
		if( $entries ) {
			echo "Total log entries: ".count($raw_log_entries)."\n";
			echo "Showing $entries entries\n";
			for($i=0; $i < $entries && isset($raw_log_entries[$i]); $i++)
				echo $raw_log_entries[$i];
		} else {
			echo date('r').": No log entries to show.\n";
		}
		echo "</pre>";
	}
	static function clear_log() {
		$ts = "[".date("r")."]\n";
		// Overwrite log file leaving permissions alone.
		@file_put_contents($ts."Log cleared".squarepay_logend_record, squarepay_log_logfile);	
	}
	static function truncate_log($entries=100) {
		$raw_log_entries = array_reverse(explode(squarepay_log_end_record, @file_get_contents(squarepay_log_logfile)));
		if( !current($raw_log_entries) )
			$raw_log_entries = array();
		$cur_log_entries = count($raw_log_entries);
		if( $cur_log_entries < $entries )
			return $cur_log_entries;
		$newlog = array();
		for($i=0; $i < $entries && isset($raw_log_entries[$i]); $i++) 
			$newlog = $raw_log_entries[$i];
		$newlog = array_reverse($newlog);
		if( $newlog )
			file_put_contents(squarepay_log_logfile, implode(squarepay_log_end_record, $newlog));
		return count($newlog);
	}
}

function squarepay_log_dump_log() {
	if( !empty(squarepay_log::$buf) ) {
		$ts = "[".date("r")."]\n";
		@file_put_contents(squarepay_log_logfile, $ts.implode("\n", squarepay_log::$buf)."\r\n", FILE_APPEND);
		if( !defined('AJAX_REQUEST') && squarepay_log::$output && squarepay_log::$log_area && (strpos(squarepay_log::$log_area, AREA) !== false) )
			echo "<pre>$ts".implode("\n", squarepay_log::$buf)."</pre>";
	}
}

function sq_log($s, $level=0) { return sq_dbg($s, false, $level);}
function sq_dbg($s, $pre=false, $level=0) {
	if( $pre ) printf("<pre>%s</pre>", htmlspecialchars($s));
	elseif( class_exists('squarepay_log') ) squarepay_log::add_msg($s, $level);
	return $s;
}

squarepay_log::init(Registry::get('addons.squarepay.debug') );

}	// class exists log
?>