<?php

/*
 * Copyright 2016, 1st Source IT, LLC, EZ Merchant Solutions
 * All rights reserved.
 * Resale prohibited.
 */
if( !defined('BOOTSTRAP') ) die('Access denied');
use Tygh\Registry;
use Tygh\Http;

// Helper functions for cs-cart

class squarepay {
	Private $access_token='';
	Private $base_url='';
	Private $version = 'v2';
	Private $http_headers = array();
	Private $order_prefix = '';
	Private $location_id = null;
	Private $auth_only = false;
	Private $currency = '';
	Private $payment_actions = array('AUTHORIZED', 'CAPTURED', 'VOIDED');
	
	Private $customer = array();
	
	Private $last_request = array();
	Private $last_response = array();
	
	Private $log_level = 1;
	Private $authorized_status ='O';
	Private $capture_status = 'P';
	Private $void_status = 'I';
	Private $refund_status = 'I';
	Private $addr_fields = array('address'=>'address_line_1', 
								'address_2'=>'address_line_2', 
								'city'=>'locality', 
								'state'=>'administrative_district_level_1', 
								'zipcode'=> 'postal_code', 
								'country'=>'country'
							);
	Public $locations = array();
	Public $location_names = array();
	Public $location_nickname = '';

	
	
	function __construct($payment_id) {
		$pp = fn_get_processor_data($payment_id);
		$token = sq_valid_access_token($payment_id);
		if(!($this->access_token = $token) ) {
			sq_log(sprintf("%s: No access token (%s)", __METHOD__, empty($pp['processor_params']['access_token']) ? "Not set" : "Expired") );
			return null;
		}
		
		$this->base_url = (sq_test_site() 
								? 'https://connect.squareupsandbox.com/' 
								: 'https://connect.squareup.com/')
							 . $this->version . "/";
		$this->http_headers = array('Authorization: Bearer ' . $this->access_token,
									'Accept: application/json'
									);
		
		$this->locations = $this->get_locations();
//sq_log(__METHOD__.": get_locations('$payment_id'):".print_r($this->locations,true) );
		$this->location_names = $this->get_location_names($this->locations);
//sq_log(__METHOD__.": location_names('$payment_id'):".print_r($this->location_names,true) );

		if( sq_test_site() ) {
sq_log("Test site is ACTIVE");
			$this->location_nickname = 'Sandbox';
			$this->location_id = SQ_SB_LOCATION_ID;
		} else {
			$this->location_id = empty($pp['processor_params']['sq_location'])
				? null
				: $pp['processor_params']['sq_location'];
//sq_log(__METHOD__.": current_location data:".print_r($this->locations[$this->location_id],true) );
			$this->location_nickname =
				empty($this->locations[$this->location_id]['name'])
					? "unknown"
					: $this->locations[$this->location_id]['name'];
		}
//sq_log("location_id=$this->location_id, location_nickname=$this->location_nickname");

		$this->currency = !empty($pp['processor_params']['sq_currency'] )
							? $pp['processor_params']['sq_currency']
							// Change from 'USD' to CART_SECONDARY_CURRENCY
							: CART_SECONDARY_CURRENCY;
		$this->order_prefix = empty($pp['processor_params']['sq_order_prefix'])
								? ''
								: $pp['processor_params']['sq_order_prefix'];
		$this->auth_only = empty($pp['processor_params']['sq_transaction_type'])
							|| $pp['processor_params']['sq_transaction_type'] == 'AUTHORIZE';

		
		$this->authorized_status = Registry::ifget('addons.squarepay.authorized_status', 'O');
		$this->capture_status = Registry::ifget('addons.squarepay.capture_status', 'P');
		$this->void_status = Registry::ifget('addons.squarepay.void_status', 'I');
		$this->refund_status = Registry::ifget('addons.squarepay.refund_status', 'I');
		
		$this->log_level = Registry::get('addons.squarepay.debug');
		
		// 8/3/16: Improved handling of unknown location nickname
		if( empty($this->location_id) ) {
			// Use first location
			$v = current($this->locations);
			$this->location_id = empty($v)
							? null	
							: $v['id'];
			$valid_names = $this->get_location_names($this->locations);
			$s = sprintf("%s: Unknown location '%s'. Valid names for access_code '%s' are: %s", 
						 __METHOD__, $this->location_nickname,
						 $this->access_token,
						$valid_names ? print_r($valid_names,true) : "None set");
			sq_log($s);
//			die($s);
		}
	}
	
	Private function get_location() {
		// 10/15/22 return this->location_id;
		return $this->location_id ? $this->location_id : false;
	}
	
	// Only used for debugging - remove at later date.
	Public function get_params() {
		return array('access_token' => $this->access_token,
					 'location_nickname' => $this->location_nickname,
					 'currency' => $this->currency,
					 'order_prefix' => $this->order_prefix,
					 'auth_only' => (integer)$this->auth_only,
					 'http_headers' => $this->http_headers,
					 'base_url' => $this->base_url,
					 'log_level' => $this->log_level,
					 'location_id' => $this->location_id
					 );
	}					 
		
	Public function get_locations() {
		$response = $this->send_request('locations', array(), 'get');
		if( $msg = $this->get_error_detail($response) ) {
			sq_log(__METHOD__.": empty response: ".$msg);
		}
		if( !empty($response['locations']) ) {
			$result = array();
			foreach($response['locations'] as $loc) {
				$result[$loc['id']] = $loc;
			}
			return $result;
		}
		return $response;
	}
	
	Public function get_location_names($locations=null) {
		$result = array();
		if( $locations ) {
			$response = $locations;
		} else {
			$response = $this->get_locations();
			if( $msg = $this->get_error_detail($response) ) {
				sq_log(__METHOD__.": $msg");
				return $result;
			}
		}
		foreach($response as $id => $r_data) {
				$result[$id] = $r_data['name'];
		}
		return $result;
	}

	
	Private function send_request($uri, $send_data=array(), $request_type='') {
		$request_type = $request_type ? $request_type 
										: ($send_data ? 'post' : 'get');
		$url = $this->base_url . $uri;
		if( $request_type != 'get')	
			sq_log(__METHOD__.": $request_type: $uri", logDetail);
		$headers = $this->http_headers;
		if( $request_type == 'put' || $request_type == 'post' )
			$headers[] = 'Content-type: application/json';
			
		// 10/4/16: For some reason, delete is different than all other http requests!!!
		switch($request_type) {
			case 'delete':
				$jdata = Http::$request_type($url, array('headers' => $headers) );
				break;
			default:
				$jdata = Http::$request_type($url, $send_data ? json_encode($send_data,true) : array(), 
											array('headers' => $headers) );
		}
		$data = json_decode($jdata, true);
		$this->last_request = array('type'=>$request_type,
									'url' => $url,
									'send_data' => $send_data,
									'headers' => $headers
									);
		$this->last_response = $data;
		if( $this->get_error_detail($data) ) {
			sq_log(__METHOD__."($url): request:".print_r($this->get_request(),true) );
			sq_log(__METHOD__."($url): response:".print_r($this->get_response(),true) );
		}
		return $data;
	}
		
	Public function get_error_detail($data=array()) {
		$errors = array();
		$data = $data ? $data : $this->last_response;
		if( !empty($data['errors']) ) {
			foreach($data['errors'] as $detail) {
				$errors[] = $detail['detail'];
			}
			return implode("\n", $errors);
		}
		return '';
	}
	
	Public function get_error_code($data=array()) {
		$errors = array();
		$data = $data ? $data : $this->last_response;
		if( !empty($data['errors']) ) {
			foreach($data['errors'] as $detail) {
				$errors[] = $detail['code'];
				break;
			}
			return implode("\n", $errors);
		}
		return '';
	}
	
	Private function generate_error($category, $code, $detail) {
		return array('errors' => array(array('category'=>$category,
									   'code'=>$code,
									   'detail'=>$detail
									   )));
	}
	
	Public function get_response() {
		return $this->last_response;
	}
	
	Public function get_request() {
		return $this->last_request;
	}
	
	Public function refund_transaction($transaction_id, $amount) {
		$uri = "refunds";
		$tx_data = array('idempotency_key' => uniqid($transaction_id),
						'payment_id' => $transaction_id,
						'reason' => 'user refund',
						'amount_money' => array(
												// Add currency
												'amount' => sq_tx_amount($amount, $this->currency),
												'currency' => $this->currency
												)
						);
		return $this->send_request($uri, $tx_data, 'post');
	}
	
	Private function get_payment_info($order_id) {
		$data = db_get_field("SELECT data FROM ?:order_data WHERE type='P' AND order_id=?i", $order_id);
		return $data ? unserialize(fn_decrypt_text($data)) : array();
	}
	
	Private function put_payment_info($order_id, $data) {
		$ud = array('data'=>fn_encrypt_text(serialize($data)),
					'order_id' => $order_id,
					'type' => 'P'
					);
		db_query("REPLACE INTO ?:order_data ?e", $ud);
	}
		
	Private function pay_to_tx_data($pay) {
		$result = array();
		foreach($pay as $k => $pd) {
			$result[] = $this->pay_to_tx($pd);
		}
		return $result;
	}
	Private function pay_to_tx($data) {
		return sq_pay_to_tx($data);
	}
	
	Private function tx_data_to_pay($tx_data) {
		$result = array();
		foreach($tx_data as $k => $tx) {
			$result[] = $this->tx_to_pay($tx);
		}
		return $result;
	}
	Private function tx_to_pay($data) {
		return sq_tx_to_pay($data);
	}
	
	Private function remove_nonce(&$order_info) {
		$pay_info = $this->get_payment_info($order_info['order_id']);	// Stored data
		unset($order_info['payment_info']['sq_nonce'], $pay_info['sq_nonce'],
			  $order_info['payment_info']['sq_save_card'], $pay_info['sq_save_card']
			);
		// SCA
		unset($order_info['payment_info']['verification_token'], $pay_info['verification_token'],
			  $order_info['payment_info']['source_id'], $pay_info['source_id']
			  );
		// SCA
		
		if( $pay_info )
			$this->put_payment_info($order_info['order_id'], $pay_info);
			
		// 3/15/17 - added removal from cart
		if( isset($_SESSION['cart']['payment_info']['sq_nonce']) )
			unset($_SESSION['cart']['payment_info']['sq_nonce']);
		// SCA
		if( isset($_SESSION['cart']['payment_info']['verification_token']) )
			unset($_SESSION['cart']['payment_info']['verification_token']);
		if( isset($_SESSION['cart']['payment_info']['source_id']) )
			unset($_SESSION['cart']['payment_info']['source_id']);
		if( isset($_SESSION['cart']['payment_info']['sq_save_card']) )
			unset($_SESSION['cart']['payment_info']['sq_save_card']);

		// SCA
	}
	
	Private function get_addr_field($section, $field, &$order_info) {
		return  !empty($order_info[$section.$field]) ? $order_info[$section.$field] : null;
	}
	
	
	/*
	 * Deprecated - use tx_refund() instead
	 */
	Public function refund_payment($order_id, $alt_amount=0) {
		trigger_error("Deprecated method called:". __METHOD__.": ".sq_backtrace());
		die("Deprecated: ".__METHOD__);
	}
	
	/*
	 * Deprecated - use tx_captue() instead
	 */
	Public function capture_payment($order_id) {
		trigger_error("Deprecated method called:". __METHOD__.": ".sq_backtrace());
		die("Deprecated: ".__METHOD__);
	}
		
	/*
	* Deprecated - use tx_void() instead.
	*/
	Public function void_payment($order_id) {
		trigger_error("Deprecated method called:". __METHOD__.": ".sq_backtrace());
		die("Deprecated: ".__METHOD__);
	}
	
	/* 
	 * New send_payment_request() works with the new verifyBuyer token and re-orders
	 * the customer/card creation to use the source-id from the payment.
	 * 1 - make payment
	 * 2 - create customer if needed
	 * 3 - save card
	 * 4 - return payment result
	 */
	Public function send_payment_request(&$order_info, $auth_only=false, $amount=0) {
		// using order_id will overwrite existing transactions
		$order_id = $order_info['order_id'];
		// 11/11/16: Added check of registry setting
		$save_card = !empty($order_info['payment_info']['sq_save_card']) 
					|| (Registry::ifget('addons.squarepay.always_save_cards', 'Y') == 'Y') ;
sq_log("order_info[payment_info]:".print_r($order_info['payment_info'],true) );
		
		$nonce = empty($order_info['payment_info']['sq_nonce']) ? '' : $order_info['payment_info']['sq_nonce'];
		// SCA
		$verification_token = empty($order_info['payment_info']['verification_token']) ? '' : $order_info['payment_info']['verification_token'];
		// SCA
		$customer_id = empty($order_info['payment_info']['sq_customer_id'])
							? false
							: $order_info['payment_info']['sq_customer_id'];
		$card_id = empty($order_info['payment_info']['sq_card_id'])
							? false
							: $order_info['payment_info']['sq_card_id'];
		$sq_data = array();
		$user_id = $order_info['user_id'];
		
		// New for card save related to verifyBuyer()
		$payment_src = '';
		
		$sq_order_id = $order_info['repaid'] ? $order_id . '_' . $order_info['repaid'] : $order_id;
		$sq_order_id = $this->order_prefix . $sq_order_id;
		$order_transaction_key = uniqid("sq_$sq_order_id");
		
		
		$tx_data = array('location_id' => $this->location_id,
						 'idempotency_key' => $order_transaction_key,
						 'source_id' => $nonce ? $nonce : $card_id,
						 'verification_token' => $verification_token,
						 'note' => sprintf("Order: %d", $order_id),
						 'amount_money' => array(
								// Add currency
								'amount' => (integer)sq_tx_amount($amount ? $amount : $order_info['total'], $this->currency),
								'currency' => $this->currency
							),
						 'autocomplete' => $auth_only ? false : true
//						 'delay_capture' => $auth_only
						 );
		// Get the shipping addresss
		$addr_data = array();
		foreach($this->addr_fields as $addr_field => $sq_addr_field) {
			$v = $this->get_addr_field('s_', $addr_field, $order_info);
			if( $v)
				$addr_data[$sq_addr_field] = $v;
		}
		if( $addr_data)
			$tx_data['shipping_address'] = $addr_data;
		$addr_data = array();
		
		if( empty($tx_data['shipping_address']) ) {
			foreach($this->addr_fields as $addr_field => $sq_addr_field) {
				$v = $this->get_addr_field('b_', $addr_field, $order_info);
				if( $v)
					$addr_data[$sq_addr_field] = $v;
			}
			if( $addr_data )
				$tx_data['shipping_address'] = $addr_data;
		}
		
		if( $nonce ) {	// Charge first
			sq_log(__METHOD__.": Using saved nonce (order_id=$order_id)", logDetail);
		} elseif( !$card_id) {
			$error = $this->generate_error('Payment', "BAD_ORDER_DATA", $msg = "No token or saved card submitted in payment info.  Please re-enter card data.");
			sq_log(__METHOD__.": $msg");
			$this->remove_nonce($order_info);
			return $error;
		}
		if( !empty($order_info['email']) )
			$tx_data['buyer_email_address'] = trim($order_info['email']);
		$tx_data['reference_id'] = $order_id;
		
		// 11/23/22 - Remove empty customer_id
		if( empty($tx_data['customer_id']) )
			unset($tx_data['customer_id']);
	
		$payment_response = $this->send_request("payments", $tx_data, 'post');

		if( $this->get_error_detail() ) {
			$this->remove_nonce($order_info);	// SCA also removes verification_token, source_id and sq_save_card
			return $payment_response;
		} else {
			$payment_src = $payment_response['payment']['id'];	// SCA: Instead of verification_token for verifyBuyer
		}
		
		// Save the transaction details to the order
		// Add the tx_data
		$tx_data = array(
			 'timestamp' => TIME,
		 	 'text_response' => __("sq_successful"),
			 'amount' => $amount ? $amount : $order_info['total'],
			 'card_brand' => empty($payment_response['payment']['card_details']['card']['card_brand'])
								? ''
								: $payment_response['payment']['card_details']['card']['card_brand'],
			'last_4' => empty($payment_response['payment']['card_details']['card']['last_4']) 
								? '' 
								: $payment_response['payment']['card_details']['card']['last_4'],
			'reason' => empty($info['reason']) ? 'New charge' : $info['reason'],
			'transaction_id' => empty($payment_response['payment']['id']) 
									? ''
									: $payment_response['payment']['id'],
			'balance' => $amount ?: $order_info['total'],
			// ON a charge, parent is this payment
			'parent_transaction_id' => empty($payment_response['payment']['id']) 
									? ''
									: $payment_response['payment']['id'],
			);
		// Since this is always the primary payment request
		// Force the tx_data to represent this.
		$_tx[$tx_data['transaction_id']][$tx_data['transaction_id']] = $tx_data;
		sq_put_tx_data($order_id, $_tx);

		
		// Now create customer and save card as needed
		if( $nonce && $user_id) {	// Not an existing card and is a cs-cart customer
			$sq_data = $this->get_user_sq_data($user_id);
			if( !$sq_data ) {	// not a square customer - create it
				sq_log("\t Adding new customer for user_id=$user_id", logDetail);
				$order_info['payment_info']['source_id'] = $payment_src;	// SCA instead of verification token
				$c_response = $this->create_customer($order_info);
				if( $c_error = $this->get_error_detail() ) {
					sq_log(sprintf("Can't create customer: user_id=%d (order_id=$order_id)$ %s", $user_id, $c_error));
				} else {
					sq_log(__METHOD__.": Create new square customer for user_id=$user_id customer_id=".$c_response['customer']['id'], logLog);
					$sq_data = array('id'=>$c_response['customer']['id']);
					// set the customer_id
					$customer_id = $sq_data['id'];
					// Given the current square defect in creating customer and card from same payment
					// We save the customer to ensure that we can then get the card on the next purchase!
					$this->put_user_sq_data($user_id, $sq_data);
sq_log("Forced save of customer_id=$sq_data[id] to user_id=$user_id");
sq_log("get_user_sq_data($user_id):".print_r($this->get_user_sq_data($user_id),true) );
				}
			} else {	// Already customer - identify it
				$customer_id = $sq_data['id'];
			}
			
			// Customer created if need be
			// Add this card.  NONCE is already used....
			if( $customer_id && $save_card ) {
				$order_info['payment_info']['source_id'] = $payment_response['payment']['id'];
				$card = $payment_response['payment']['card_details']['card'];
			
				$card_response = $this->add_customer_card($customer_id, $order_info, $card );
				if( $card_error = $this->get_error_detail($card_response) ) {
					sq_log(sprintf("Can't add card to customer id '%s' (order_id=$order_id):%s", $customer_id, $card_error));
					$this->remove_nonce($order_info);
					// Probably shouldn't return the failure of the card since the payment is already done.
//					return $card_response;
				} else {
					sq_log(__METHOD__.": Added card for user_id=$user_id from order_id=$order_id", logLog);
					// set the card_id
					$card_id = $card_response['card']['id'];
					$sq_data['cards'][$card_id] = $card_response['card'];
					
					// Save the user's square data
					$this->put_user_sq_data($user_id, $sq_data);
				}
			} elseif($customer_id) {
				sq_log("\t Skip adding card for new customer", logDetail);
			}
		}	// If nonce && user_id
		return $payment_response;
	}
	
	Private function mkdate($ts) {
		return date("Y-m-d", $ts) . "T" . date("H:i:s", $ts) .date("P", $ts);
	}
	
	// Testing only
	Public function list_transactions($begin_ts, $end_ts, $sort_order='DESC', $cursor='') {
		$begin_date = $this->mkdate($begin_ts);
		$end_date = $this->mkdate($end_ts);
		$uri = sprintf("locations/%s/transactions?begin_time=%s&end_time=%s&sort_order=%s%s",
					   	urlencode($this->location_id), urlencode($begin_date), urlencode($end_date),
						urlencode($sort_order), $cursor ? ("&cursor=".urlencode($cursor)) : '');
		return $this->send_request($uri, array(), 'get');
	}
	
	
	Public function get_customer($customer_id) {
		$uri = "customers/".urlencode($customer_id);
		return $this->customer = $this->send_request($uri, array(), 'get');
	}
	
	// Also uses the order_info structure for $cust_data
	Public function create_customer($info) {
		$user_id = $info['user_id'];
		if( $this->get_user_sq_data($user_id) ) {
			sq_log("User ($user_id) already exists", logLog);
			return $this->generate_error('Customer', 'CUSTOMER_EXISTS', "Customer (user_id=$user_id) already exists");
		}
		$uri = 'customers';
		$sq_user_id = $this->order_prefix . $user_id;
		$idempotency_key = uniqid("sq_cust_$sq_user_id");
		$tx_data = array( 'idempotency_key' => $idempotency_key,
						  'given_name' => !empty($info['firstname']) ? $info['firstname'] 
										: ($info['b_firstname'] ? $info['b_firstname'] : ''),
						  'family_name' => !empty($info['lastname']) ? $info['lastname'] 
										: ($info['b_lastname'] ? $info['b_lastname'] : ''),
						// 3/22/17 - added trim to email address since cart seems to not clean the data.
						  'email_address' => empty($info['email']) ? '' : trim($info['email']),
						  'phone' => empty($info['phone']) ? '' : $info['phone'],
						  'reference_id' => $user_id,
						  'source_id' => $info['payment_info']['source_id'],
						  'note' => ''
						  );
		
		// Get the address info if it exists
		$addr_data = array();
		foreach($this->addr_fields as $addr_field => $sq_addr_field) {
			$v = $this->get_addr_field('b_', $addr_field, $order_info);
			if( $v)
				$addr_data[$sq_addr_field] = $v;
		}
		if( $addr_data )
			$tx_data['address'] = $addr_data;

		$response = $this->send_request($uri, $tx_data, 'post');
		
		if( !$this->get_error_detail() ) {
			// update the user_data
			sq_log(__METHOD__.": Created customer ($user_id)", logLog);
			$sq_data = array('id'=>$response['customer']['id']);
			$this->put_user_sq_data($user_id, $sq_data);
		}
		return $response;
	}
	
	// Card info is 'card_nonce' and 'billing_address'
	// We use the order_info data map as template.
	Public function add_customer_card($customer_id, $info, $card) {
		$card['customer_id'] = $customer_id;
		$customer = $this->get_customer($customer_id);
		$order_id = empty($info['order_id']) ? 0 : $info['order_id'];
		$user_id = empty($info['user_id']) ? 0 : $info['user_id'];
		if( $this->get_error_detail() ) {
			sq_log("Error retrieving customer:".print_r($customer,true));
			$this->remove_nonce($info);
			return $this->generate_error('Customer', 'NO_SUCH_CUSTOMER', "Customer ($customer_id) does not exist");
		}

		$uri = 'cards';
		$name = sprintf("%s %s", !empty($info['b_firstname']) ? $info['b_firstname'] 
										: (!empty($info['firstname']) ? $info['firstname'] : ''),
								! empty($info['b_lastname']) ? !empty($info['b_lastname']) 
										: (!empty($info['lastname']) ? $info['lastname'] : '') 
						);
		$card['cardholder_name'] = $name;
		
		// Use the nonce or sourece_id to set the card
		if( !empty($info['payment_info']['source_id']) ) {
			$tx_data = array('source_id'=> $info['payment_info']['source_id'] );
		} elseif( !empty($info['payment_info']['sq_nonce']) ) {
			$nonce = $info['payment_info']['sq_nonce'];
			$tx_data = array('card_nonce' => $nonce );
		}
		$tx_data['card'] = $card;
		
		$user_id = $this->order_prefix . $info['user_id'];
		$idempotency_key = uniqid("sq_card_$user_id");

		$tx_data['idempotency_key'] = $idempotency_key;
						 
		// Get the address info if it exists
		$addr_data = array();
		foreach($this->addr_fields as $addr_field => $sq_addr_field) {
			$v = $this->get_addr_field('b_', $addr_field, $order_info);
			if( $v)
				$addr_data[$sq_addr_field] = $v;
		}

		if( $addr_data)
			$tx_data['billing_address'] = $addr_data;

		// Note: Email no longer a supported field in card API
		$result = $this->send_request($uri, $tx_data, 'post');
		
		if( !$this->get_error_detail() ) {
			// Update the customer data.
			sq_log(__METHOD__.": Added card for user_id=$user_id from order_id=$order_id", logLog);
			$user_id = $info['user_id'];
			$sq_data = $this->get_user_sq_data($user_id);
			$sq_data['cards'][$result['card']['id']] = $result['card']; 
			$this->put_user_sq_data($user_id, $sq_data);
		}		
		return $result;
	}
	
	Public function delete_customer($user_id, $customer_id) {
		$uri = "customers/".urlencode($customer_id);
		$response = $this->send_request($uri, array(), 'delete');
		if( $error = !$this->get_error_detail() ) {
			sq_log(__METHOD__.": Deleted customer_id=$customer_id", logLog);
			$this->put_user_sq_data($user_id, array());
		} else {
			sq_log(__METHOD__.": Error deleting customer_id=$customer_id:".print_r($error,true));
		}
		return $error ? false : true;
	}
	
	Public function update_customer($customer_id, $info) {
		$user_id = $info['user_id'];
		$sq_data = $this->get_user_sq_data($user_id);
		$uri = 'customers';
		$cards = array();
		foreach($sq_data as $k => $v) {
			switch($k) {
				case 'address':
					foreach($sq_data[$k] as $k2 => $v2) {
						if( isset($info[$k][$k2]) )
							$sq_data[$k][$k2] = $info[$k][$k2];
					}
					break;
				case 'cards':
					$cards = $sq_data['cards'];
					unset($sq_data['cards']);
					break;
				default:
					if( isset($info[$k]) )
						$sq_data[$k] = $info[$k];
					break;
			}
		}
		$response = $this->send_request($uri, $tx_data, 'put');
		
		if( !$this->get_error_detail() ) {
			// update the user_data
			sq_log(__METHOD__.": Updated customer (user_id=$user_id) (customer_id=$customer_id)", logLog);
			$sq_data = array('id'=>$response['customer']['id']);
			if( $cards )
				$sq_data['cards'] = $cards;
			$this->put_user_sq_data($user_id, $sq_data);
		}
		return $response;
	}
	
	function delete_customer_card($user_id, $customer_id, $card_id) {
		// Get the user data
		$sq_data = $this->get_user_sq_data($user_id);
		if( empty($sq_data) || empty($sq_data['cards'][$card_id]) ) {
			sq_log("Delete customer card: No such card for user_id=$user_id");
			return $this->generate_error('Customer', 'NO_CARD', "No such card for user (user_id=$user_id");
		}
		$uri = "customers/".urlencode($customer_id)."/cards/".urlencode($card_id);
		$uri = "cards/".urlencode($card_id)."/disable";
		$response = $this->send_request($uri, array(), 'post');
		
		// Not sure why the INVALID_REQUEST_ERROR is being generated for card removals
		// But can only assume that we simply won't list/use those cards in the future
		if( !$this->get_error_detail() || $response['errors'][0]['category'] == 'INVALID_REQUEST_ERROR') {
			if( $this->get_error_detail($response) )	// log the error if it exists
				sq_log(__METHOD__.": Can't delete square card. (customer_id=$customer_id, card_id=$card_id)");
			// But remove it from the user's data
			unset($sq_data['cards'][$card_id]);
			$this->put_user_sq_data($user_id, $sq_data);
			sq_log(__METHOD__.": Deleted customer card ($card_id) for user_id=$user_id", logLog);
			return array();		// Don't return the actual error.
		} else {
			if( $response['errors'][0]['category'] != 'INVALID_REQUEST_ERROR' )
				sq_log(__METHOD__.": Error: ".$this->get_error_detail($response) );
		}
		return $response;
	}


	// Cart support
	Public function get_user_sq_data($user_id) {
		$data = db_get_field("SELECT sq_data FROM ?:users WHERE user_id=?i", $user_id);
		return $data ? unserialize($data) : array();
	}
	
	Public function put_user_sq_data($user_id, $sq_data) {
sq_log(__METHOD__.": user_id=$user_id, sq_data1:".print_r($sq_data,true) );
		$sq_data = $this->de_dup_sq_data($sq_data);
sq_log(__METHOD__.": user_id=$user_id, sq_data2 (after de_dup):".print_r($sq_data,true) );
		db_query("UPDATE ?:users SET sq_data=?s WHERE user_id=?i", serialize($sq_data), $user_id);
	}
		
	// Extra TX functions
	Public function tx_charge(&$info, $auth_only=false, $amount=0) {
		// using order_id will overwrite existing transactions
		$order_id = empty($info['order_id']) ? 0 : $info['order_id'];
		
		$nonce = empty($info['payment_info']['sq_nonce']) ? '' : $info['payment_info']['sq_nonce'];
		// SCA
		$verification_token = empty($info['payment_info']['verification_token']) ? '' : $info['payment_info']['verification_token'];
		// SCA
		$customer_id = empty($info['payment_info']['sq_customer_id'])
							? false
							: $info['payment_info']['sq_customer_id'];
		$card_id = empty($info['payment_info']['sq_card_id'])
							? false
							: $info['payment_info']['sq_card_id'];
		
		// Process the payment
		$sq_order_id = $info['repaid'] ? $order_id . '_' . $info['repaid'] : $order_id;
		$sq_order_id = $this->order_prefix . $sq_order_id;
		$order_transaction_key = uniqid("sq_$sq_order_id");
		
		$tx_data = array('idempotency_key' => $order_transaction_key);
		// SCA
		if( $verification_token )
			$tx_data['verification_token'] = $verification_token;
		// SCA
		
		// Get the address info if it exists
		$addr_data = array();
		foreach($this->addr_fields as $addr_field => $sq_addr_field) {
			$v = $this->get_addr_field('s_', $addr_field, $order_info);
			if( $v)
				$addr_data[$sq_addr_field] = $v;
		}
		if( !empty($addr_data))
			$tx_data['shipping_address'] = $addr_data;
		
		// Get the address info if it exists
		$addr_data = array();
		foreach($this->addr_fields as $addr_field => $sq_addr_field) {
			$v = $this->get_addr_field('b_', $addr_field, $order_info);
			if( $v)
				$addr_data[$sq_addr_field] = $v;
		}
		if( !empty($addr_data))
			$tx_data['billing_address'] = $addr_data;
			
		$tx_data = array_merge( $tx_data, array(
/*
						 'billing_address' => array(
								'address_line_1' => empty($info['b_address']) ? '' : $info['b_address'],
								'address_line_2' => empty($info['b_address_2']) ? '' : $info['b_address_2'],
								'locality' => empty($info['b_city']) ? '' : $info['b_city'],
								'administrative_district_level_1' => empty($info['b_state']) ? '' : $info['b_state'],
								'postal_code' => empty($info['b_zipcode']) ? '' : trim($info['b_zipcode']),
								'country' => empty($info['b_country']) ? '' : $info['b_country']
								),
*/
						 'amount_money' => array(
								 // Add currency
								 'amount' => (integer)sq_tx_amount($amount ? $amount : $info['total'], $this->currency),
								 'currency' => $this->currency
								 ),
						 'autocomplete' => $auth_only ? false : true
//						 'delay_capture' => $auth_only
						 ));
		if( !empty($info['reason']) )
			$tx_data['note'] = sprintf("Reason: %s", $info['reason']);
		
		$tx_data['source_id'] = empty($info['payment_info']['sq_nonce'])
									? ( !empty($customer_id) && !empty($card_id)
										? $card_id
										: ''
									  )
									: $nonce;
		if( empty($tx_data['source_id']) ) {
			// Must either have nonce or customer/card_id
			$error = $this->generate_error('Payment', "BAD_ORDER_DATA", "<pre>No customer_id, card_id or nonce in payment info:".print_r($info['payment_info'],true) );
			return $error;
		}
		$tx_data['verification_token'] = $verification_token;
		$tx_data['idempotency_key'] = uniqid($this->order_prefix.$order_id);
		
		// 3/22/17 - added trim to email address since cart seems to not clean the data.
		if( !empty($info['email']) )
			$tx_data['buyer_email_address'] = trim($info['email']);
		$tx_data['reference_id'] = $order_id;
	
		$response = $this->send_request("payments", $tx_data, 'post');
//sq_log(__METHOD__.": response:".print_r($response,true), logDetail);
		
		if( !$this->get_error_detail($response) ) {
			// Create the new extra tx record
			$tx_data = sq_get_tx_data($order_id);
			$tx_data = array('action' => $auth_only ? 'AUTHORIZED' : 'CAPTURED',
							 'timestamp' => TIME,
						 	 'text_response' => __("sq_successful"),
							 'amount' => $amount ? $amount : $info['total'],
							 'card_brand' => empty($response['payment']['card_details']['card']['card_brand'])
												? ''
												: $response['payment']['card_details']['card']['card_brand'],
							'last_4' => empty($response['payment']['card_details']['card']['last_4']) 
												? '' 
												: $response['payment']['card_details']['card']['last_4'],
							'reason' => empty($info['reason']) ? 'New charge' : $info['reason'],
							'transaction_id' => empty($response['payment']['id']) 
													? ''
													: $response['payment']['id'],
							'balance' => $amount ?: $info['total'],
							// ON a charge, parent is this transaction
							'parent_transaction_id' => empty($response['payment']['id']) 
													? ''
													: $response['payment']['id'],
//							'tender_id' => empty($response['transaction']['tenders'][0]['id']) 
//												? ''
//												: $response['transaction']['tenders'][0]['id']
							);
			$this->update_tx_data($order_id, $tx_data['transaction_id'], $tx_data);
			sq_log(sprintf("%s: order_id=%s, Charged '%.2f'", __METHOD__, $order_id, fn_format_price($amount)));
		}	// if not error			
		return $response;
	}

	Public function tx_refund($order_id, $parent_transaction_id, $amount, $change_status=true) {
		list($tx_chain, $cur_tx) = sq_find_extra_tx($order_id, $parent_transaction_id, 'CAPTURED');
		if( empty($cur_tx) ) {
			return $this->generate_error('Refund', 'MISSING_TRANSACTION_ID', 
										 sprintf("Can't find corresponding %s transaction_id=%s", 
													'CAPTURED', $parent_transaction_id) 
										);
		}
		
		$working_balance = fn_format_price($tx_chain[$parent_transaction_id][$parent_transaction_id]['amount']);
		if( isset($cur_tx['balance']) )
			$working_balance = fn_format_price($cur_tx['balance']);
		elseif( isset($tx_chain[$parent_transaction_id][$parent_transaction_id]['balance']))
			$working_balance = fn_format_price($tx_chain[$parent_transaction_id][$parent_transaction_id]['balance']);

		$pre_balance = fn_format_price($working_balance - $amount);
		
		if( $pre_balance  < 0 ) {	// Not enough funds for the request
			return $this->generate_error('Refund', 'INSUFFICENT_FUNDS', 
										 sprintf("Request for %.2f exceeds available balance of %.2f", 
													$$working_balance, $pre_balance) 
										);
		}
		// Perform the traansaction
		$refund = $this->refund_transaction($parent_transaction_id, fn_format_price($amount));
		$refund_value = fn_format_price($cur_tx['amount'] * -1);
		if( !$this->get_error_detail($refund) ) {
			$refunded = fn_format_price($refund['refund']['amount_money']['amount'] /100) * -1;		// get actual
			$avail_balance = fn_format_price($cur_tx['amount'] + $refunded);	// compute from actual
			$tx_data = array('action'=>'REFUNDED',
							 'timestamp' => TIME,
							 'text_response' => __("sq_successful"),
							 'transaction_id'=> uniqid('refund'),
							 'parent_transaction_id' => $parent_transaction_id,
							 'balance' => $avail_balance,
//							 'tender_id' => $tender_id,
						 	 'text_response' => __("sq_successful"),
							 'amount'=> $refunded
							 );
			$this->update_tx_data($order_id, $parent_transaction_id, $tx_data);
			$cur_status = db_get_field("SELECT status from ?:orders WHERE order_id=?i", $order_id);
			if( $change_status) {
				fn_change_order_status($order_id, $this->refund_status, $cur_status);
				sq_log(sprintf("%s: order_id=%s, refunded \$%.2f, order status changed to %s", __METHOD__, $order_id, fn_format_price($amount), $this->refund_status), logDetail);
			} else {
				sq_log(sprintf("%s: order_id=%s, refunded \$%.2f, order status not changed", __METHOD__, $order_id, fn_format_price($amount)), logDetail);
			}

		}
		return $refund;
	}

	Public function tx_void($order_id, $parent_transaction_id) {
		list($tx_chain, $cur_tx) = sq_find_extra_tx($order_id, $parent_transaction_id, 'AUTHORIZED');
//		$cur_tx = sq_find_extra_tx($order_id, $parent_transaction_id,'AUTHORIZED');
		if( empty($cur_tx) ) {
			return $this->generate_error('Void', 'MISSING_TRANSACTION_ID', 
										 sprintf("Can't find corresponding %s transaction_id=%s", 'AUTHORIZED', $parent_transaction_id) );
		}
		$amount = ($cur_tx['amount'] * -1);
		$uri = "payments/$parent_transaction_id/cancel";
		$void =  $this->send_request($uri, array(), 'post');
		if( !$this->get_error_detail($void) ) {
			$tx_data = array('action'=>'VOIDED',
							 'timestamp' => TIME,
						 	 'text_response' => __("sq_successful"),
							 'amount' => $amount,
							 'transaction_id'=> 'void',
							 'parent_transaction_id' => $parent_transaction_id,
							 'balance' => 0,
							 'tender_id' => empty($void['tender_id']) ? $cur_tx['tender_id'] : $void['tender_id'],
						 	 'text_response' => __("sq_successful"),
							 );
			$this->update_tx_data($order_id, $parent_transaction_id, $tx_data);
			sq_change_payment_action($order_id, 'VOIDED', $this->void_status);
		}
		return $void;
	}
	
	Public function tx_capture($order_id, $parent_transaction_id) {
		list($tx_chain, $cur_tx) = sq_find_extra_tx($order_id, $parent_transaction_id, 'AUTHORIZED');
//		$cur_tx = sq_find_extra_tx($order_id, $transaction_id, 'AUTHORIZED');
		if( empty($cur_tx) || !isset($tx_chain[$parent_transaction_id][$parent_transaction_id]) ) {
			return $this->generate_error('Void', 'MISSING_TRANSACTION_ID', 
										 sprintf("Can't find corresponding %s transaction_id=%s", 'AUTHORIZED', $parent_transaction_id) );
		}
		$parent =& $tx_chain[$parent_transaction_id][$parent_transaction_id];
		$uri = sprintf("payments/%s/complete", $parent_transaction_id);
		$capture = $this->send_request($uri, array(), 'post');
		if( !$this->get_error_detail($capture) ) {
			$tx_data = array('action'=>'CAPTURED',
							 'timestamp' => TIME,
						 	 'text_response' => __("sq_successful"),
							 'transaction_id'=> "captured",
							 'parent_transaction_id' => $parent_transaction_id,
							 'balance' => isset($cur_tx['balance']) ? $cur_tx['balance'] : $cur_tx['amount'],
//							 'tender_id' => $cur_tx['tender_id'],
						 	 'text_response' => __("sq_successful"),
							 'amount' => 0	// Nothing changed so amount is zero
							 );
			$this->update_tx_data($order_id, $parent_transaction_id, $tx_data);
			sq_change_payment_action($order_id, 'CAPTURED', $this->capture_status);
		}
		return $capture;
	}

	Private function update_tx_data($order_id, $parent_transaction_id, $tx) {
		$tx_data = sq_get_order_transactions($order_id);	// Get all
		$hit = false;
		
		$type = empty($tx_data[$parent_transaction_id][$tx['transaction_id']])
					? 'creating'
					: 'updating';
		$tx_data[$parent_transaction_id][$tx['transaction_id']] = $tx;
		sq_put_tx_data($order_id, $tx_data);
		return;
	}

	 /*
	 * Use for non-order payments
	 * returns array(customer_data, card_data, user_data, transactio_rdata)
	 * If error, returns array( category, code, detail)
	 */
	Public function single_pay($info, $auth_only=false, $amount=0 ) {
return $this->generate_error('Deprecated', __FUNCTION__, 'Function no longer supported');	     

	    $customer_data = $card_data = $transaction_data = $sq_data = array();
	     
		$save_card = !empty($info['payment_info']['sq_save_card']) 
					|| (Registry::ifget('addons.squarepay.always_save_cards', 'Y') == 'Y') ;
		
		$nonce = empty($info['payment_info']['sq_nonce']) ? null : $info['payment_info']['sq_nonce'];
		// SCA
		$verification_token = empty($info['payment_info']['verification_token']) ? null : $info['payment_info']['verification_token'];
		// SCA
		$customer_id = empty($info['payment_info']['sq_customer_id'])
							? null
							: $info['payment_info']['sq_customer_id'];
		$card_id = empty($info['payment_info']['sq_card_id'])
							? null
							: $info['payment_info']['sq_card_id'];
		$sq_data = array();
		$user_id = isset($info['user_id']) ? $info['user_id'] : ( empty($auth['user_id']) ? null : $auth['user_id]']) ;
		
		// Anonymous not allowed
		if( empty($user_id))
		    return $this->generate_error('Users', 'NO_USER', 'No user_id');
		    
		$sq_data = $this->get_user_sq_data($user_id);
		$customer_id ?: (empty($sq_data['id']) ? null : $sq_data['id']);
//sub_log(__METHOD__.": customer_id=$customer_id, sq_data:".print_r($sq_data,true).", info:".print_r($info,true));
		
		//
		// Customer_data
		//
		if( empty($customer_id) ) {
		    $c_response = $this->create_customer($info);
		    if( $c_error = $this->get_error_detail())
		        return $c_response;
		    $customer_id = $sq_data['id'] = $c_response['customer']['id'];
		    $customer_data = $c_response['customer'];
		} else {
		    // Get existing customer info
		    $c_response = $this->get_customer($customer_id);
		    if( $c_error = $this->get_error_detail())
		        return $c_response;
		    $customer_id = $sq_data['id'] = $c_response['customer']['id'];
		    $customer_data = $c_response['customer'];
		}

		//
		// Card data - nonce == new card, !nonce == existing card_data
		//
		// Customer already created and info is in $customer_data
		//
		if( $nonce ) {
		    // Using new card or replacing existing card_data
		    $card_response = $this->add_customer_card($customer_id, $info);
		    if( $card_error = $this->get_error_detail($card_response) )
		        return $card_response;
		    $card_id = $card_response['card']['id'];
		    $card_data = $card_response['card'];
			$nonce =  $info['payment_info']['sq_nonce'] = null;
		} elseif($card_id) {
			$valid_card = false;
			if( !empty($customer_data['cards'])) {	// verify card in place
				foreach($customer_data['cards'] as $k => $v) {
					if( $v['id'] == $card_id ) {
						$valid_card = $k;
						break;
					}
				}
			}
		    if( $valid_card === false){
		        $error = $this->generate_error('Card', 'BAD_CARD_ID', "Card id ($card_id) not in customer data");
		        sq_log(__METHOD__.": ".$this->get_error_detail($error));
		        return $error;
		    }
		    $card_data = $customer_data['cards'][$valid_card];
		} else {
		    $error = $this->generate_error('Card', 'BAD_CARD_DATA', "Missing nonce or card_id");
		    sq_log(__METHOD__.": ".$this->get_error_detail($error));
		    return $error;
		}
		if( !empty($customer_data['cards'])) {
			foreach($customer_data['cards'] as $k => $v) {
				if( $k == $valid_card )
					continue;
				$sq_data['cards'][$v['id']] = $v;
			}
		}
		
        // Reapply info for transaction
		$info['payment_info']['sq_card_id'] = $card_id;
		$info['payment_info']['sq_customer_id'] = $customer_id;
		$info['payment_info']['save_card'] = true;
		
		
		// Process the payment
		$idepotency_key = uniqid($customer_id);
		$note = empty($info['note']) ? "Generic transaction" : $info['note'];
		

		// 11/23/16: Added note to tx_data.
		$tx_data = array('idempotency_key' => $idepotency_key,
						 'note' => $note,
						 'amount_money' => array(
								// Change to use sq_tx_amount with currency
								'amount' => sq_tx_amount($amount, $this->currency),
								'currency' => $this->currency
						 	),
						 'autocomplete' => $auth_only ? false : true
//						 'delay_capture' => $auth_only
						 );
		// SCA
		if( $verification_token )
			$tx_data['verificationToken'] = $verification_token;
		// SCA
		
		// Get the shipping addresss
		$addr_data = array();
		foreach($this->addr_fields as $addr_field => $sq_addr_field) {
			$v = $this->get_addr_field('s_', $addr_field, $info);
			if( $v)
				$addr_data[$sq_addr_field] = $v;
		}
		if( $addr_data)
			$tx_data['shipping_address'] = $addr_data;
		$addr_data = array();
		
		if( empty($tx_data['shipping_address']) ) {
			foreach($this->addr_fields as $addr_field => $sq_addr_field) {
				$v = $this->get_addr_field('b_', $addr_field, $info);
				if( $v)
					$addr_data[$sq_addr_field] = $v;
			}
			if( $addr_data )
				$tx_data['shipping_address'] = $addr_data;
		}
		
		if( $customer_id && $card_id ) {	// Existing customer with card on file or new card
			$tx_data['customer_id'] = $customer_id;
			$tx_data['customer_card_id'] = $card_id;
			sq_log(__METHOD__.": Using customer/card_id", logDetail);
		} elseif( $nonce ) {
			$tx_data['card_nonce'] = $nonce;
			sq_log(__METHOD__.": Using saved nonce", logDetail);
		} else {
			// Must either have nonce or customer/card_id
			$error = $this->generate_error('Payment', "BAD_DATA", $msg = "No token submitted in payment info.  Please re-enter card data.");
			sq_log(__METHOD__.": ".$this->get_error_detail($error));
			return $error;
		}
		
		// 3/22/17 - added trim to email address since cart seems to not clean the data.
		if( !empty($info['email']) )
			$tx_data['buyer_email_address'] = trim($info['email']);
		$tx_data['reference_id'] = empty($info['reference_id']) ? (string)time() : $info['reference_id'];

		if( $amount >= 1 ) {	// Square only supports charges of $1 or more
			$response = $this->send_request("locations/$this->location_id/transactions", $tx_data, 'post');
			if( $this->get_error_detail() )
				return $response;
			$transaction_data = $response['transaction'];
		}

		$sq_data['cards'][$card_id] = $card_data;	// Add the card
		$sq_data = $this->de_dup_sq_data($sq_data);	// remove any duplcate last_4 based on expiry

		return array(
		            $customer_data,
		            $card_data,
		            $sq_data,
		            $transaction_data
		            );
	}
	
Public function de_dup_sq_data($sq_data) {
	$dups = array();
	if( !empty($sq_data['cards'])) {	
		foreach($sq_data['cards'] as $cid => $c_info) {
			$dups[$c_info['last_4']][$cid] = $c_info;
		}
		foreach($dups as $last_4 => $dup_data) {
			$exp_val = 0;
			// Multople instances of this card (last 4).
			foreach($dup_data as $cid => $v) {
				if( empty($cid)) {	// no card cid in array - bad data
					unset($sq_data['cards'][$cid]);
					unset($dup_data[$last_4][$cid]);
					continue;
				}
				if( $exp_val >= (integer)$v['exp_year'].$v['exp_month'] ) {	// dup found, delete it
					unset($sq_data['cards'][$cid]);
					unset($dup_data[$last_4][$cid]);
				} else {
					$exp_val = (integer)$v['exp_year'].$v['exp_month'];
				}
			}
		}
	}
	return $sq_data;
}
}	// end class squarepay
?>