<?php
if( !defined('BOOTSTRAP') ) die('Access denied');

use Tygh\Registry;
use Tygh\Settings;

// Compatibility
$dir_addons = Registry::get('config.dir.addons');
$index_script = Registry::get('config.admin_index');

$v = explode(DIRECTORY_SEPARATOR, __FILE__);
$addon_name = $v[count($v)-4];

/* Every addon must have this to bootstrap the ez_common addon
 */
if( file_exists($f=ez_fix_path($dir_addons."ez_common/lib/upgrade.php")) ) {
	require_once($f);
} else {	// Need to force update of the tools
	require_once(ez_fix_path($dir_addons."$controller/lib/upgrade/upgrade.php"));		// copy of common version of tools - used to bootstrap the ez_common addon
}


$default_payment_id=15;
$v = explode(DIRECTORY_SEPARATOR, __FILE__);
$addon_name = $v[count($v)-4];

switch($mode) {
	case 'log':
		// 7/21/17 - fixed class name
		$log_class="squarepay_log";
		if( class_exists('ez_log') ) {
			$log_lines = 10;
			$l_buf = array();
			switch($action) {
				case 'clear': 
					$log_class::clear_log(); 
					fn_set_notification('N', $controller, "Cleared log file");
					break;
				case 'truncate':
					$lines = (empty($_REQUEST['truncate_lines']) && $dispatch_extra) 
								? $dispatch_extra 
								: (empty($_REQUEST['truncate_lines']) ? 100 : $_REQUEST['truncate_lines']);
					$entries = $log_class::truncate_log($lines);
					fn_set_notification('N', $controller, "Truncated log to $entries lines");
					break;
				case 'show':
					$lines = min($log_lines, (empty($_REQUEST['show_lines']) && $dispatch_extra) 
								? $dispatch_extra 
								: (empty($_REQUEST['show_lines']) ? 100 : $_REQUEST['show_lines']));
					$log_class::show_log($lines);
					exit;
			}	// switch
		}	// class exists
		break;
	
		
	case 'refund':
		if( $_SERVER['REQUEST_METHOD'] == 'POST' ) {
			$order_id = empty($_REQUEST['order_id']) ? 0 : $_REQUEST['order_id'];
			if( empty($order_id) ) {
				fn_set_notification("E", __("error"), "No order_id for $mode", 'K');
				return array(CONTROLLER_STATUS_OK, "orders.manage");
			}
			$refund_amount = $_REQUEST['refund_amount'];
			if( $refund_amount <= 0.0 ) {
				fn_set_notification("E", __("error"), "Bad refund amount '$refund_amount'", 'K');
				return array(CONTROLLER_STATUS_OK, "orders.details&amp;order_id=$order_id");
			}
			$payment_id = sq_order_payment_id($order_id);
			$sqObj = new squarepay($payment_id);
			$payment_info = sq_get_payment_info($order_id);
			$transaction_id = $payment_info['transaction_id'];
			$result = $sqObj->tx_refund($order_id, $transaction_id, $refund_amount);
			if( $error = $sqObj->get_error_detail($result) ) {
				fn_set_notification("E", __("error"), $error, 'K');
			} else {
				fn_set_notification("N", __("notice"), "Refunded order_id $order_id in amount of $refund_amount", 'K');
			}
			fn_redirect("orders.details&amp;order_id=$order_id");
		}
		break;
	case 'capture':
		$order_id = empty($_REQUEST['order_id']) ? 0 : $_REQUEST['order_id'];
		if( empty($order_id) ) {
			fn_set_notification("E", __("error"), "No order_id for $mode", 'K');
			return array(CONTROLLER_STATUS_OK, "orders.manage");
		}
		$payment_id = sq_order_payment_id($order_id);
		$payment_info = sq_get_payment_info($order_id);
		$transaction_id = $payment_info['transaction_id'];
		$sqObj = new squarepay($payment_id);
		$result = $sqObj->tx_capture($order_id, $transaction_id);
		if( $error = $sqObj->get_error_detail($result) ) {
			fn_set_notification("E", __("error"), $error, 'K');
		} else {
			fn_set_notification("N", __("notice"), "Captured payment for order_id $order_id", 'K');
		}
		fn_redirect("orders.details&amp;order_id=$order_id");
	case 'void':
		$order_id = empty($_REQUEST['order_id']) ? 0 : $_REQUEST['order_id'];
		if( empty($order_id) ) {
			fn_set_notification("E", __("error"), "No order_id for $mode", 'K');
			return array(CONTROLLER_STATUS_OK, "orders.manage");
		}
		$payment_id = sq_order_payment_id($order_id);
		$payment_info = sq_get_payment_info($order_id);
		$transaction_id = $payment_info['transaction_id'];
		$sqObj = new squarepay($payment_id);
		$result = $sqObj->tx_void($order_id, $transaction_id);
		if( $error = $sqObj->get_error_detail($result) ) {
			fn_set_notification("E", __("error"), $error, 'K');
		} else {
			fn_set_notification("N", __("notice"), "Voided payment for order_id $order_id", 'K');
		}
		fn_redirect("orders.details&amp;order_id=$order_id");
		
	case 'tx_charge':
		if( empty($_REQUEST['order_info']['order_id']) )
			die("$controller.$mode: no order_id");
		$order_id = $_REQUEST['order_info']['order_id'];
		$order_info = fn_get_order_info($order_id);
		if( empty($order_info) )
			die("$controller.$mode: no order info");
		$order_info = array_merge($order_info, $_REQUEST['order_info']);
		$payment_id = $order_info['payment_id'];
		$pp = fn_get_processor_data($payment_id);
		if( empty($pp['processor_params']) )
			die("$controller.$mode: no processor data");
		$params = $pp['processor_params'];
		$auth_only = $params['sq_transaction_type'] == 'AUTHORIZE';
		$sqObj = new squarepay($payment_id);
		$response = $sqObj->tx_charge($order_info, $auth_only);
		if( $error = $sqObj->get_error_detail($response) ) {
			fn_set_notification("E", __("error"), $error, 'K');
		} else {
			$msg = str_replace('%[AMOUNT]%', sq_amount_to_currency($order_info['total']), __("sq_successful_charge_text") );
			fn_set_notification('N', __("notice"), $msg);
		}
		fn_redirect("orders.details&amp;order_id=$order_id");
		
	case 'tx_capture':
		$order_id = empty($_REQUEST['order_id']) ? 0 : $_REQUEST['order_id'];
		$transaction_id = empty($_REQUEST['transaction_id']) ? 0 : $_REQUEST['transaction_id'];
		if( empty($order_id) ) {
			fn_set_notification("E", __("error"), "No order_id for $mode", 'K');
			return array(CONTROLLER_STATUS_OK, "orders.details&amp;order_id=$order_id");
		}
		if( empty($transaction_id) ) {
			fn_set_notification("E", __("error"), "No transaction_id for $mode", 'K');
			return array(CONTROLLER_STATUS_OK, "orders.details&amp;order_id=$order_id");
		}
		$payment_id = sq_order_payment_id($order_id);
		$sqObj = new squarepay($payment_id);
		$result = $sqObj->tx_capture($order_id, $transaction_id);
		if( $error = $sqObj->get_error_detail($result) ) {
			fn_set_notification("E", __("error"), $error, 'K');
		} else {
			fn_set_notification("N", __("notice"), "Captured payment for transaction_id $transaction_id", 'K');
		}
		fn_redirect("orders.details&amp;order_id=$order_id");
		
	case 'tx_void':
		$order_id = empty($_REQUEST['order_id']) ? 0 : $_REQUEST['order_id'];
		$transaction_id = empty($_REQUEST['transaction_id']) ? 0 : $_REQUEST['transaction_id'];
		if( empty($order_id) ) {
			fn_set_notification("E", __("error"), "No order_id for $mode", 'K');
			return array(CONTROLLER_STATUS_OK, "orders.details&amp;order_id=$order_id");
		}
		if( empty($transaction_id) ) {
			fn_set_notification("E", __("error"), "No transaction_id for $mode", 'K');
			return array(CONTROLLER_STATUS_OK, "orders.details&amp;order_id=$order_id");
		}
		$payment_id = sq_order_payment_id($order_id);
		$sqObj = new squarepay($payment_id);
		$result = $sqObj->tx_void($order_id, $transaction_id);
		if( $error = $sqObj->get_error_detail($result) ) {
			fn_set_notification("E", __("error"), $error, 'K');
		} else {
			fn_set_notification("N", __("notice"), "Voided payment for transaction_id $transaction_id", 'K');
		}
		fn_redirect("orders.details&amp;order_id=$order_id");
		
	case 'tx_refund':
		$order_id = empty($_REQUEST['order_id']) ? 0 : $_REQUEST['order_id'];
		$transaction_id = empty($_REQUEST['transaction_id']) ? 0 : $_REQUEST['transaction_id'];
		if( empty($order_id) ) {
			fn_set_notification("E", __("error"), "No order_id for $mode", 'K');
			return array(CONTROLLER_STATUS_REDIRECT, "orders.details&amp;order_id=$order_id");
		}
		if( empty($transaction_id) ) {
			fn_set_notification("E", __("error"), "No transaction_id for $mode", 'K');
			return array(CONTROLLER_STATUS_REDIRECT, "orders.details&amp;order_id=$order_id");
		}
		$refund_amount = empty($_REQUEST['refund_amount']) ? 0: $_REQUEST['refund_amount'];
		if( empty($refund_amount) ) {
			fn_set_notification("E", __("error"), "No refund data for $mode", 'K');
			return array(CONTROLLER_STATUS_REDIRECT, "orders.details&amp;order_id=$order_id");
		}
//sq_log("refund_amount=$refund_amount");
		$payment_id = sq_order_payment_id($order_id);
		$sqObj = new squarepay($payment_id);
		$result = $sqObj->tx_refund($order_id, $transaction_id, $refund_amount);
//sq_log(__FUNCTION__.":result:".print_r($result,true));
		if( $error = $sqObj->get_error_detail($result) ) {
			fn_set_notification("E", __("error"), $error, 'K');
		} else {
			fn_set_notification("N", __("notice"), "Refunded payment for transaction_id $transaction_id", 'K');
		}
		fn_redirect("orders.details&amp;order_id=$order_id");	
	
	case 'delete_card':
		$user_id = $_REQUEST['user_id'];
		$card_id = $_REQUEST['card_id'];
		$customer_id = $_REQUEST['customer_id'];
		$payment_id = $_REQUEST['payment_id'];
		$order_id = $_REQUEST['order_id'];
		
		if(empty($user_id) || empty($card_id) || empty($customer_id) || empty($payment_id) || empty($order_id)) {
			sq_log("Admin: Bad user/card_id/customer_id/payment_id (user_id=$user_id, card_id=$card_id, customer_id=$customer_id, payment_id=$payment_id)");
			fn_set_notification("E", __("error"), "Bad user/card_id/customer_id/payment_id/order_id", 'K');
			fn_redirect("orders.details&amp;order_id=$order_id");
		}
		$sqObj = new squarepay($payment_id);
		$response = $sqObj->delete_customer_card($user_id, $customer_id, $card_id);
		// 10/4/16: Fixed bug where $this was used instead of $sqObj
		if( !$response || !$sqObj->get_error_detail($response) ) {
			fn_set_notification('N', __("notice"), __("sq_deleted_card"));
		} else {
			fn_set_notification('E', __("error"), $sqObj->get_error_detail($response), 'K');
		}
		fn_redirect("orders.details&amp;order_id=$order_id");
		
	case 'reprocess_order':
		$order_id = $action ? $action : (empty($_REQUEST['order_id']) ? 0 : $_REQUEST['order_id']);
		if( empty($order_id) )
			die("No order_id");
		$pay_info = db_get_field("SELECT data FROM ?:order_data WHERE type='P' AND order_id=?i", $order_id);
		if( $pay_info ) 
			$pay_info = unserialize(fn_decrypt_text($pay_info));
		if( empty($pay_info['sq_nonce']) )
			die("No token available for reprocess");
		fn_start_payment($order_id);
		die("Done");
		
	case 'install_lang_vars':
		$force = $action ? true : false;
		sq_install_lang_vars($force);
		fn_redirect("payments.manage");
		
	case 'check_install':
	// 1/28/17: added verbose and install_templates to squarepay_install_setup() params.
		if( function_exists($f = sprintf("%s_install_setup", $addon_name)) )
			$f(array('force'=>true, 'verbose'=>true, 'install_templates'=>true));
		if( file_exists($f = Registry::get('config.dir.addons')."$addon_name/upgrade/post_upgrade_all_versions.php") )
			include($f);
		if( function_exists('addon_install_setup') ) {
			$result = addon_install_setup($controller);
			if( $result === false )
				$result = "Failed";
			else $result = sprintf("Okay: Current version=%s, next upgrade=%s", 
								   (empty($result['cur_ver'])?"not set":$result['cur_ver']),
								   (empty($result['next_upgrade'])?"next login":("after ".date("m/d/Y H:i:s",$result['next_upgrade'])))
								   );
			fn_set_notification($result == "Failed" ? 'E' : 'N', "$controller.$mode", $result, 'K');
		} else {	// function exists
			fn_set_notification('E', __("error"), "$mode not supported or ez_common not installed.", 'K');
		}
		fn_redirect("payments.manage");	
		break;
		
	case 'get_token':	// ajax request for token
		if( empty($_REQUEST['payment_id']) )
			die("Bad payment_id");
			
		$payment_id = $_REQUEST['payment_id'];
		die(sq_valid_access_token($payment_id, $log=false));
		
	case 'json_params':
		if( empty($_REQUEST['payment_id']) )
			die("Bad payment_id");
		$payment_id = $_REQUEST['payment_id'];
		$token = sq_valid_access_token($payment_id, false);	// Forces check of 'token_expires' and renews if necessary
		$params = array();
		if( $token ) {
			$p_info = fn_get_payment_method_data($payment_id);
			$params = empty($p_info['processor_params']) ? array() : $p_info['processor_params'];
			foreach($params as $k => $v) {
				if( empty($v) )
					unset($params[$k]);
			}
		}
		die(json_encode($params));
	
	case 'json_locations':
		$payment_id = empty($_REQUEST['payment_id']) ? 0 : $_REQUEST['payment_id'];
		if( empty($payment_id) )
			die("No payment_id");
		$location_names = sq_location_names($payment_id);
		die( json_encode(array('names'=>$location_names)) );
		
		if( !$location_names )
			return "<p>".__("sq_no_locations_found")."</p>";
		$s = "<p>".__("sq_locations_for_this_account")." <ul>";
		foreach($location_names as $name) {
			$s .= "<li>$name</li>";
		}
		$s .= "</ul></p>";
		echo $s;
		exit;
		
	case 'version':
		$ez_common = 'ez_common';
		addon_init($ez_common);
		$ez_version = addon_current_version($ez_common);
		if( empty($ez_version) )
			$ez_version = "not installed or missconfigured";
		addon_init($controller);
		$version = addon_current_version($controller);
		$msg = PRODUCT_NAME.": ".PRODUCT_VERSION." ".PRODUCT_EDITION."<br/>EZ Common: $ez_version<br/>$controller: $version";
		fn_set_notification('N', "$controller.$mode", $msg, true);
		fn_redirect("addons.manage");
		break;
		
	case 'test':
	case 'tests':
		if( file_exists($f=dirname(__FILE__)."/tests.php") )
			include($f);
		break;
		
	case 'no_auth_response':
		fn_set_notification('E', __("error"), __('sq_no_auth_response'), 'K');
		fn_redirect("payments.manage");
	
	default:
		if( file_exists($f = ez_fix_path($dir_addons."ez_common/lib/std_addon_controller.php")) )
			include($f);
		break;
}

$redirect = (isset($_REQUEST['c_url']) ? $_REQUEST['c_url']
				: (isset($_REQUEST['return_url']) ? $_REQUEST['return_url']
					: (isset($_REQUEST['redirect_url']) ? $_REQUEST['redirect_url']
						: Registry::get('config.customer_index')
					   )
				   )
			);
return array(CONTROLLER_STATUS_REDIRECT, $redirect);
				