<?php
if (!defined('BOOTSTRAP')) { die('Access denied'); }

use Tygh\Http;
use Tygh\Registry;

$processor_params = $processor_data['processor_params'];


if( defined('PAYMENT_NOTIFICATION') ) {
	die("No callbacks for squarepay");
} else {
	
	$payment_id = $order_info['payment_id'];
	$sqObj = new squarepay($payment_id);
	$auth_only = $processor_params['sq_transaction_type'] == 'AUTHORIZE';
	sq_log(basename(__FILE__).": payment_id=$payment_id, order_id=$order_info[order_id], auth_only=$auth_only", logDetail);
	$response = $sqObj->send_payment_request($order_info, $auth_only);

	$error_message = '';
	// 4/19/18: Changes to support no status changes on actions
	if( $auth_only ) {
		$order_status = ($order_status = Registry::get('addons.squarepay.authoirze_status'))
											? $order_status
											: 'O';
	} else {
		$order_status = ($order_status = Registry::get('addons.squarepay.capture_status'))
											? $order_status
											: 'P';
	}
//	$order_status = $auth_only ? Registry::ifget('addons.squarepay.authorize_status', 'O') 
//								: Registry::ifget('addons.squarepay.capture_status', 'P');
	if( !empty($response['errors']) ) {
		$error_code = $sqObj->get_error_detail();
		switch($error_code) {
			case 'CARD_DECLINED':
			case 'CARD_EXPIRED':
			case 'INVALID_EXPIRATION':
			case 'INVALID_EXPIRATION_YEAR':
			case 'INVALID_EXPIRATION_DATE':
			case 'INVALID_CARD':
			case 'INVALID_CARD_DATA':
			case 'VERIFY_CVV_FAILURE':
			case 'VERIFY_AVS_FAILURE':
			case 'UNSUPPORTED_CARD_BRAND':
				$order_status = 'D';
				break;
			default: 
				$order_status = 'F';
				break;
		}
		$error_message = $sqObj->get_error_detail($response);
		sq_log(basename(__FILE__).": error_message=$error_message", logDetail);
		// 11/24/16: Added notification
		fn_set_notification('E', __("error"), $error_message, 'K');
		// 0320/17 - added more clearing of various payment info on errors.
		if( !empty($_SESSION['cart']['payment_info']) ) {
			$fields = array('sq_save_cart', 'sq_card_id', 'transaction_id', 'tender_id', 'sq_nonce', 'sq_save_card', 'verification_token', 'source_id');
			foreach($fields as $field) {
				unset($_SESSION['cart']['payment_info'][$field], $_SESSION['cart']['extra_payment_info'][$field]);
				unset($order_info['payment_info'][$field], $order_info['extra_payment_info'][$field]);
			}
		}
	} else {
		/* Had a situation reported where no errors and no data.  This is an attempt to prevent a successful transaction 
		 * if there are no errors AND no transaction_id.
		 */
		$transaction_id = !empty($response['payment']['id']) ? $response['payment']['id'] : '';
		if( !$transaction_id ) {
			$order_status = 'F';
			$error_message = __("sq_no_txid");
			fn_set_notification('E', __("error"), $error_message, 'K');
			sq_log("%s: %s: HTTP error:%s\nresponse:%s", basename(__FILE__), $error_message, Http::getError(), print_r($response,true));
			if( !empty($_SESSION['cart']['payment_info']) ) {
				$fields = array('sq_save_cart', 'sq_card_id', 'transaction_id', 'tender_id', 'sq_nonce', 'verification_token', 'source_id', 'sq_save_card');
				foreach($fields as $field) {
					unset($_SESSION['cart']['payment_info'][$field], $_SESSION['cart']['extra_payment_info'][$field]);
					unset($order_info['payment_info'][$field], $order_info['extra_payment_info'][$field]);
				}
			}
		}
	}
		
	$pp_response = array();
	
	// 11/27/16: Remove any nonce so it can't be reused.
	if( $order_status == 'D' || $order_status == 'F' ) {
		$p_data = sq_get_payment_info($order_info['order_id']);
		unset($p_data['sq_nonce'], $order_info['payment_info']['sq_nonce'],
			  $p_data['sq_save_card'], $order_info['payment_info']['sq_save_card'],
			  $p_data['verification_token'], $order_info['payment_info']['verfication_token'],
			  $p_data['source_id'], $order_info['payment_info']['source_id'] 
			 );	// Remove the nonce so it can't be reused.
		sq_put_payment_info($order_info['order_id'], $p_data);
	}
	
	$card_brand = empty($response['payment']['card_details']['card']['card_brand'])
					? ''
					: $response['payment']['card_details']['card']['card_brand'];
	$last_4 = empty($response['payment']['card_details']['card']['last_4']) ? '' : $response['payment']['card_details']['card']['last_4'];
	
	$text_response = $error_message 
									? $error_message 
									: __("sq_successful");
	$transaction_id = !empty($response['payment']['id']) ? $response['payment']['id'] : '';
	// Needed for refund/void
	// No more tender_id
//	$tender_id = !empty($response['transaction']['tenders'][0]['id']) ? $response['transaction']['tenders'][0]['id'] : '';
	$pay_status = !empty($response['payment']['card_details']['status'])
								? $response['payment']['card_details']['status']
								: '';
	// Goal is to have format of payment area in orders.details look like:
	// Authorized: $amount
	// Mastercard: **** 1234
	// Transaction ID: $tx_id
	// Tender_id: $tender_id
	// Response: Text response from provider (reason_text)
		$pp_response[$pay_status] = sq_amount_to_currency($order_info['total']);
		$pp_response[$card_brand] = sprintf("**** %d", $last_4);
		$pp_response['transaction_id'] = $transaction_id;
//		$pp_response['tender_id'] = $tender_id;
		$pp_response['text_response'] = $text_response;
		
	$pp_response['order_status'] = $order_status;
	$tx_data = sq_pay_to_tx($pp_response);
	$transaction_id = isset($tx_data['transaction_id'])
						? $tx_data['transaction_id']
						: '';
	$cur_tx = sq_get_tx_data($order_info['order_id']);
	$cur_tx[$transaction_id][$transaction_id] =
		array_merge(empty($cur_tx[$transaction_id][$transaction_id])
						? array()
						: $cur_tx[$transaction_id][$transaction_id]
					, $tx_data);
	sq_put_tx_data($order_info['order_id'], $cur_tx);
	fn_set_hook('squarepay_payment', $pp_response, $response);
sq_debugging() && sq_log("request:".print_r($sqObj->get_request(), true).
						"provider response:".print_r($response,true).
						"pp_response:".print_r($pp_response,true), logDetail );		
}